<?php

namespace Drupal\proc\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultAllowed;
use Drupal\Core\Access\AccessResultForbidden;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Path\CurrentPathStack;
use Drupal\Core\Session\AccountProxy;
use Drupal\proc\Entity\Proc;
use Drupal\proc\Traits\ProcRecipientTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Controller.
 */
class ProcController extends ControllerBase  {

  use ProcRecipientTrait;

  /**
   * The current user.
   *
   * @var AccountProxy
   */
  protected AccountProxy $current_user;

  /**
   * The current path.
   *
   * @var CurrentPathStack
   */
  protected CurrentPathStack $current_path;

  /**
   * ProcController controller.
   *
   * @param AccountProxy $current_user
   *   The file repository.
   *
   * @param CurrentPathStack $current_path
   *   The current path.
   *
   */
  public function __construct(AccountProxy $current_user, CurrentPathStack $current_path) {
    $this->currentUser = $current_user;
    $this->currentPath = $current_path;
  }

  /**
   * {@inheritdoc}
   *
   * @param ContainerInterface $container
   *   The Drupal service container.
   *
   * @return static
   */
  public static function create(ContainerInterface $container): static {
    return new static (
      $container->get('current_user'),
      $container->get('path.current')
    );
  }

  /**
   * Checks whether the user is a possible recipient for a PGP message.
   *
   * @param string $users
   *   User IDs as CSV.
   *
   * @return AccessResultForbidden
   *   Controller response.
   */
  public function checkRecipientsCsvController($users) {
    $rejected_recipient_candidate = $this->checkRecipientsCsv($users);
    $encryption_strategy = $this->config('proc.settings')->get('proc-encryption-strategy') ?? 0;
    switch ($encryption_strategy) {
      case 0: // Restrictive strategy
        if (!is_int($rejected_recipient_candidate) && $rejected_recipient_candidate != FALSE) {
          return AccessResult::allowed();
        }
        return AccessResult::forbidden(
          $this->t('User account %rejected_recipient_candidate does not exist or user is not a holder of PGP keys.', [
            '%rejected_recipient_candidate' => $rejected_recipient_candidate,
        ]));
        break;

      case 1: // Permissive strategy
        return AccessResult::allowed();
        break;
    }
  }

  /**
   * Checks whether the current user is listed as recipient of proc entity.
   */
  public function checkCurrentUserIsRecipient(): AccessResultForbidden|AccessResultAllowed {
    // Get the current user.
    $current_user = $this->currentUser->id();
    // Get proc ID from current path.
    $current_path = $this->currentPath->getPath();

    if ($current_path == '/proc/generate-keys') {
      return AccessResult::allowed();
    }

    $proc_id = str_replace('/proc/', '', $current_path);
    if (!is_numeric($proc_id)) {
      // Edit page:
      $proc_id = substr($proc_id, 0, strpos($proc_id, "/"));
      // Grant access if user is author or admin.
      if (is_numeric($proc_id)) {
        $proc = Proc::load($proc_id);
        if (
          ($proc->getOwnerId() == $current_user && $this->currentUser->hasPermission('edit proc entity')) ||
          $this->currentUser->hasPermission('administer proc entity')
        ) {
          return AccessResult::allowed();
        }
      }
    }
    // Load proc entity by proc ID.
    $proc = Proc::load($proc_id);
    if ($proc instanceof Proc) {
      // Get the recipients of the proc entity.
      $recipients = $proc->get('field_recipients_set')->getValue();

      foreach ($recipients as $recipient) {
        if ($recipient['target_id'] == $current_user) {
          return AccessResult::allowed();
        }
      }
    }

    return AccessResult::forbidden();
  }
}
