<?php

namespace Drupal\proc\Form;

use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\InvokeCommand;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\File\FileSystem;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Render\Markup;
use Drupal\Core\Url;
use Drupal\file\FileRepository;
use Drupal\proc\Entity\Proc;
use Drupal\proc\Traits\ProcRecipientTrait;
use Drupal\user\UserStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Drupal\Core\Session\AccountProxy;

/**
 * Encrypt content.
 */
class ProcEncryptForm extends FormBase {

  use ProcRecipientTrait;

  const TEXTAREA = '1';
  const TEXTFIELD = '2';

  /**
   * The file system.
   *
   * @var FileSystem
   */
  protected FileSystem $file_system;

  /**
   * The file repository.
   *
   * @var FileSystem
   */
  protected FileRepository $file_repository;

  /**
   * The current user.
   *
   * @var AccountProxy
   */
  protected AccountProxy $current_user;

  /**
   * The user storage.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected $userStorage;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * ProcEncryptForm form.
   *
   * @param FileSystem $file_system
   *   The file system.
   *
   * @param FileRepository $file_repository
   *  The file repository.
   *
   * @param AccountProxy $current_user
   *   The file repository.
   *
   * @param \Drupal\user\UserStorageInterface $user_storage
   *   The user storage.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(FileSystem $file_system, FileRepository $file_repository, AccountProxy $current_user, UserStorageInterface $user_storage, ConfigFactoryInterface $config_factory) {
    $this->fileSystem = $file_system;
    $this->fileRepository = $file_repository;
    $this->currentUser = $current_user;
    $this->userStorage = $user_storage;
    $this->configFactory = $config_factory;
  }

  /**
   * {@inheritdoc}
   *
   * @param ContainerInterface $container
   *   The Drupal service container.
   *
   * @return static
   */
  public static function create(ContainerInterface $container): static {
    return new static (
      $container->get('file_system'),
      $container->get('file.repository'),
      $container->get('current_user'),
      $container->get('entity_type.manager')->getStorage('user'),
      $container->get('config.factory')
    );
  }


  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'proc_encrypt_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $hidden_fields = [
      'cipher_text',
      'source_file_name',
      'source_file_size',
      'source_file_type',
      'source_file_last_change',
      'browser_fingerprint',
      'generation_timestamp',
      'generation_timespan',
      'signed',
    ];

    foreach ($hidden_fields as $hidden_field) {
      $form[$hidden_field] = ['#type' => 'hidden'];
    }

    // Get query string:
    $query = $this->getRequest()->query->all();

    $field_input_mode = FALSE;
    // Get the input mode:
    if (isset($query['field-input-mode'])) {
      $field_input_mode = $query['field-input-mode'];
    }

    // File is the default input mode:
    if (!$field_input_mode) {
      $form['proc-file'] = [
        '#type' => 'file',
        '#description' => $this->t('Select a file for encryption.'),
        'data' => $query,
      ];
    }

    $field_type = '';
    if ($field_input_mode === self::TEXTAREA) $field_type = 'textarea';
    if ($field_input_mode === self::TEXTFIELD) $field_type = 'textfield';

    $form['proc-text-field'] = [
      '#type' => $field_type,
      '#description' => $this->t('Write a text for encryption.'),
    ];

    // Get current URL. If it has a query string standalone=FALSE, then we
    // do not use ajax submit.
    $proc_standalone_mode = FALSE;
    if (isset($query['proc_standalone_mode'])) {
      $proc_standalone_mode = $query['proc_standalone_mode'];
    }

    // If stand-alone mode is false, then this encryption
    // is triggered by a field widget and the dialog has to
    // be closed by ajax submit.
    if ($proc_standalone_mode == 'FALSE') {
      $form['#attached']['drupalSettings'] = [
        'proc' => [
          'proc_standalone_mode' => FALSE,
        ],
      ];

      $form['submit-proc'] = [
        '#type' => 'submit',
        '#value' => $this->t('Submit'),
        '#ajax' => [
          // Close dialog using ajax submit:
          'callback' => '::submitFormAjax',
          'event' => 'click',
        ],
      ];
    }
    else {
      $form['#attached']['drupalSettings'] = [
        'proc' => [
          'proc_standalone_mode' => TRUE,
        ],
      ];

      // Do not use ajax submit once this is not triggered by a field widget
      // and therefore there is no dialog to close:
      $form['actions']['submit-proc'] = [
        '#type' => 'submit',
        '#value' => $this->t('Submit'),
      ];
    }
    return $form;
  }

  /**
   * Submit handler for the ajax submit.
   */
  public function submitFormAjax(array &$form, FormStateInterface $form_state) {
    // Close the dialog:
    $response = new AjaxResponse();
    // Selector for the dialog close button.
    $selector = '.ui-dialog-titlebar-close';
    $method = 'click';
    $arguments = [];
    $response->addCommand(new InvokeCommand($selector, $method, $arguments));

    $file_name = (isset($form['proc-file']['#value']))
      ? $form['proc-file']['#value']["\0Symfony\\Component\\HttpFoundation\\File\\UploadedFile\0originalName"]
      : '';

    // Get the latest proc ID authored by current user:
    $current_user_id = $this->currentUser->id();


    $query = \Drupal::entityQuery('proc');
    $query->accessCheck(TRUE);
    $query->condition('user_id', $current_user_id);
    $query->sort('id', 'DESC');
    $query->range(0, 1);
    $entity_ids = $query->execute();
    $proc_id = array_shift($entity_ids);

    $proc_object_label = Proc::load($proc_id)->get('label')->getValue()[0]['value'];

    if (!empty($file_name) && $file_name == $proc_object_label) {
      // Set the file name in the form:
      $field_name = $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_field_name'];
      $delta = $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_field_delta'];
      $selector = 'input[name="' . $field_name  . '[' . $delta . '][target_id]"]';
      $file_name .= ' (' . $proc_id . ')';
    }
    if (empty($file_name)) {
      $selector = '#' . $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_form_triggering_field'];


      // Text field encryption mode uses a hidden textfield as input mode and therefore accepts only the cipher text ID:
      $file_name .= $proc_id;
    }

    $response->addCommand(new InvokeCommand($selector, 'val', [$file_name]));

    return $response;
  }

  /**
   * Implements form validation.
   *
   * The validateForm method is the default method called to validate input on
   * a form.
   *
   * @param array $form
   *   The render array of the currently built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // @todo Add validation for the syntax of armored cipher text.
  }

  /**
   * Implements a form submit handler.
   *
   * The submitForm method is the default method called for any submit elements.
   *
   * @param array $form
   *   The render array of the currently built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   * @throws EntityStorageException
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    global $base_url;

    $recipients_set_ids = $form_state->get('storage');

    $destination = FALSE;
    $current_url = \Drupal::request()->headers->get('referer');
    $parse_result = UrlHelper::parse($current_url);

    if (isset($parse_result['query']['destination'])) {
      $destination = $parse_result['query']['destination'];
    }

    $cipher = ['cipher' => $form_state->getValue('cipher_text')];
    $meta = [
      'source_file_name' => $form_state->getValue('source_file_name'),
      'source_file_size' => $form_state->getValue('source_file_size'),
      'source_file_type' => $form_state->getValue('source_file_type'),
      'source_file_last_change' => $form_state->getValue('source_file_last_change'),
      'browser_fingerprint' => $form_state->getValue('browser_fingerprint'),
      'generation_timestamp' => $form_state->getValue('generation_timestamp'),
      'generation_timespan' => $form_state->getValue('generation_timespan'),
      'signed' => $form_state->getValue('signed'),
    ];

    $proc = Proc::create();

    $recipient_users = [];
    $rejected_users = [];

    switch ($this->configFactory->get('proc.settings')->get('proc-encryption-strategy')) {
      case 0: // Restrictive strategy
        foreach ($recipients_set_ids as $recipient_id) {
          $recipient_users[] = ['target_id' => $recipient_id];
        }
        break;

      case 1: // Permissive strategy
        foreach ($recipients_set_ids as $recipient_id) {
          if ($this->userHasKeyring($recipient_id)) {
            $recipient_users[] = ['target_id' => $recipient_id];
          }
          else {
            $rejected_users[] = $recipient_id;
          }
        }
        break;
    }

    $json_content = $cipher['cipher'];
    $fragment = Crypt::hashBase64(Crypt::randomBytesBase64(32));
    $json_filename = $fragment . '.json';

    $config = \Drupal::config('proc.settings');
    $enable_stream_wrapper = $config->get('proc-enable-stream-wrapper');
    $stream_wrapper = $config->get('proc-stream-wrapper');
    $block_size = $config->get('proc-file-block-size');
    $blocks_split_enabled = $config->get('proc-enable-block-size');

    $file_id = FALSE;
    if ($enable_stream_wrapper === 1 && !empty($stream_wrapper)) {
      $json_dest = $stream_wrapper;
      if (!is_dir($json_dest)) {
        $this->fileSystem->mkdir($json_dest, NULL, TRUE);
      }

      if ($json_content) {
        if ($blocks_split_enabled && !empty($block_size)) {
          // $content_lines_number = substr_count($json_content, "\n");
          $lines = explode("\n", $json_content);
          $content_lines_number = count($lines);

          $lines_size_ratio = $content_lines_number / $block_size;
          $blocks = intval($lines_size_ratio);

          $remaining = $content_lines_number % $block_size;
          if ($remaining > 0) {
            $blocks++;
          }
          $blocks_index = 0;

          $blocks_lines = [];
          $content_line_index = 0;
          while ($blocks_index < $blocks) {
            $line_in_block_index = 0;
            while ($line_in_block_index < $block_size) {
              $blocks_lines[$blocks_index][] = $lines[$content_line_index];
              $content_line_index++;
              $line_in_block_index++;
            }
            $blocks_index++;
          }

          $blocks_texts = [];
          foreach ($blocks_lines as $block_index => $block_lines) {
            foreach ($block_lines as $block_line) {
              $blocks_texts[$block_index] = $blocks_texts[$block_index] . "\n" . $block_line;
            }
          }

          $json_fids = [];
          foreach ($blocks_texts as $block_text) {
            $fragment = Crypt::hashBase64(Crypt::randomBytesBase64(32));
            $json_filename = $fragment . '.json';
            $jsonFid = $this->fileRepository->writeData(
              $block_text,
              "$json_dest/$json_filename",
              FileSystemInterface::EXISTS_REPLACE
            );
            if ($jsonFid->id()) {
              $json_fids[] = $jsonFid->id();
            }
          }
        }
        else {
          $jsonFid = $this->fileRepository->writeData(
            $json_content,
            "$json_dest/$json_filename",
            FileSystemInterface::EXISTS_REPLACE
          );
          if ($jsonFid->id()) {
            $file_id = $jsonFid->id();
          }
        }
      }
    }

    if ($file_id) {
      $cipher = ['cipher_fid' => $file_id];
    }
    if (!empty($json_fids)) {
      $cipher = ['cipher_fid' => $json_fids];
    }

    $proc->set('armored', $cipher)
      ->set('meta', $meta)
      ->set('langcode', 'en')
      ->set('label', $meta['source_file_name'])
      ->set('type', 'cipher')
      ->set('field_recipients_set', $recipient_users)
      ->save();

    $proc_id = $proc->id();
    if (is_numeric($proc_id)) {
      $url = Url::fromUri('internal:/proc/' . $proc_id);
      $link = Link::fromTextAndUrl($base_url . '/proc/' . $proc_id, $url)
        ->toString()
        ->getGeneratedLink();

      $this->messenger()->addMessage(
        $this->t(
          'Encryption is completed. Decryption link: %link', ['%link' => Markup::create($link)]
        )
      );

      if (!empty($rejected_users)) {
        $user_names = [];
        if (count($rejected_users) > 1) {
          $user_names = array_map(function($uid) {
            return $this->userStorage
              ->load($uid)
              ->getDisplayName() . ' (uid: ' . $uid . ')';
          }, $rejected_users);
          $rejected_users = implode(', ', $user_names);
        }
        else {
          $rejected_users =  $this->userStorage
            ->load(current($rejected_users))
            ->getDisplayName() . ' (uid: ' . current($rejected_users) . ')';
        }
        $this->logger('proc')
          ->warning('Recipients list of a protected content (id: @proc_id) was impacted due the lack of valid PGP key found for following user(s): @rejected_users.', [
            '@proc_id' => $proc_id,
            '@rejected_users' => $rejected_users,
        ]);
      }
    }
    else {
      $this->messenger()->addMessage($this->t('Error'), type: TYPE_ERROR);
    }
    if ($destination) {
      $url = Url::fromUri('internal:/' . $destination);
      $response = new RedirectResponse($url->toString());
      $response->send();
    }
  }

}
