<?php

namespace Drupal\proc\Plugin\Field\FieldWidget;

use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldWidget\EntityReferenceAutocompleteWidget;
use Drupal\Core\Form\FormStateInterface;
use Drupal\proc\Element\ProcField;
use Drupal\proc\Event\ProcElementEvent;
use Drupal\proc\Event\ProcEvents;

/**
 * Defines the 'proc_entity_reference_widget' field widget.
 *
 * @FieldWidget(
 *   id = "proc_entity_reference_widget",
 *   module = "proc",
 *   label = @Translation("Proc Entity Reference Field Widget"),
 *   field_types = {
 *     "proc_entity_reference_field"
 *   }
 * )
 */
class ProcEntityReferenceWidget extends EntityReferenceAutocompleteWidget {

  /**
   * Recipients fetcher endpoint URL.
   *
   * @var string
   */
  private $recipients_fetcher;

  /**
   * Form element IDs to be used as parameters for the recipients endpoint.
   *
   * @var string
   */
  private $fetcher_filter;

  /**
   * Type of identifier to be used as parameters for the recipients endpoint.
   *
   * 0: ID
   * 1: Name
   *
   * @var int
   */
  private $fetcher_filter_type;

  /**
   * Comma separated list of user IDs to be set as recipients for the encryption.
   *
   * @var string
   */
  private $default_recipients;

  /**
   * Machine name of a field containing user IDs of recipients.
   *
   * @var int
   */
  private $recipients_to;

  /**
   * Machine name of a field containing user IDs of CC recipients.
   *
   * @var int
   */
  private $recipients_cc;

  /**
   * Mode of operation.
   *
   * @var int
   */
  private $op_mode;

  /**
   * Modes of data input.
   *
   * 0: Disabled
   * 1: Only encryption
   * 2: Encryption and signature
   *
   * @var int
   */
  private $input_mode;

  /**
   * Modes of decryption.
   *
   * 0: New page
   * 1: Dialog
   *
   * @var int
   */
  private $decryption_mode;

  /**
   * Password caching flag.
   *
   * 0: Disable password caching
   * 1: 'Show checkbox for allowing password caching
   * 2: 'Hide the checkbox and apply password caching by default
   *
   * @var int
   */
  private $pass_caching;

  /**
   * Trigger format.
   *
   * 0: Encrypt and Decrypt buttons
   * 1: Toggle checkbox (check for encryption, uncheck for decryption)
   *
   * @var int
   */
  private $trigger_format;

  /**
   * Form submit element ID for the master field.
   *
   * @var string
   */
  private $master_field_submit_id;

  /**
   * Flag for triggering decryption of another encrypted field.
   *
   * @var int
   */
  private $trigger_decryption;

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
        'proc_field_recipients_fetcher_endpoint' => '',
        'proc_field_recipients_fetcher_endpoint_filter' => '',
        'proc_field_recipients_fetcher_endpoint_filter_type' => '',
        'proc_field_recipients_manual_fetcher' => '',
        'proc_field_recipients_to_field' => '',
        'proc_field_recipients_cc_field' => '',
        'proc_field_mode' => 1,
        'proc_field_input_mode' => 0,
        'proc_field_decryption_mode' => '',
        'proc_allow_password_caching' => '',
        'proc_trigger_format' => '',
        'proc_master_field_submit_id' => '',
        'proc_decryption_trigger_fields' => '',
      ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element[] = parent::settingsForm($form, $form_state);

    $element['proc_field_recipients_to_field'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Machine name of the field containing the user IDs of recipients'),
      '#description' => $this->t('It must be a user reference field visible in the same form as this field.'),
      '#default_value' => $this->getSetting('proc_field_recipients_to_field') ?? '',
    ];

    $element['proc_field_recipients_cc_field'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Machine name of the field containing user IDs of CC recipients'),
      '#description' => $this->t('It must be a user reference field visible in the same form as this field.'),
      '#default_value' => $this->getSetting('proc_field_recipients_cc_field') ?? '',
    ];

    $element['proc_field_recipients_manual_fetcher'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Define the user IDs of recipients in a CSV list'),
      '#description' => $this->t('Example: 1,2,3'),
      '#default_value' => $this->getSetting('proc_field_recipients_manual_fetcher') ?? '',
      '#disabled' => FALSE,
    ];

    $element['proc_field_recipients_fetcher_endpoint'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Endpoint for fetching user IDs of recipients'),
      '#description' => $this->t('Example: https://example.com/api/get/users/param-a/param-b/param-c?param-d=1&#38;param-e=2'),
      '#default_value' => $this->getSetting('proc_field_recipients_fetcher_endpoint') ?? '',
      '#disabled' => FALSE,
    ];

    $element['proc_field_recipients_fetcher_endpoint_token_tree'] = [
      '#theme' => 'token_tree_link',
      '#token_types' => ['site', 'current-page'],
      '#show_restricted' => TRUE,
    ];

    $element['proc_field_recipients_fetcher_endpoint_filter'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Identifier of form elements to be used as parameters for the endpoint'),
      '#description' => $this->t('Example: edit-field-a,edit-field-b,edit-field-c'),
      '#default_value' => $this->getSetting('proc_field_recipients_fetcher_endpoint_filter') ?? '',
      '#disabled' => FALSE,
    ];

    $element['proc_field_recipients_fetcher_endpoint_filter_type'] = [
      '#type' => 'radios',
      '#title' => $this->t('Type of identifier'),
      '#default_value' => $this->getSetting('proc_field_recipients_fetcher_endpoint_filter_type') ?? 0,
      '#options' => $this->getFetcherEndpointFilterTypes(),
      '#disabled' => FALSE,
    ];

    $element['proc_field_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Modes of operation'),
      '#default_value' => $this->getSetting('proc_field_mode') ?? 0,
      '#options' => $this->getFieldOperationModes(),
      2 => [
        '#disabled' => TRUE, // Encryption and signature.
      ],
    ];

    $element['proc_field_input_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Modes of input'),
      '#default_value' => $this->getSetting('proc_field_input_mode') ?? '',
      '#options' => $this->getInputModes(),
      1 => [
        '#disabled' => TRUE, // Textarea field.
      ],
      3 => [
        '#disabled' => TRUE, // Date field.
      ],
    ];

    $element['proc_field_decryption_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Modes of decryption'),
      '#default_value' => $this->getSetting('proc_field_decryption_mode') ?? '',
      '#options' => $this->getDecryptionModes(),
    ];

    $element['proc_allow_password_caching'] = [
      '#type' => 'radios',
      '#title' => $this->t('Password caching'),
      '#default_value' => $this->getSetting('proc_allow_password_caching') ?? '',
      '#options' => $this->getPasswordCachingOptions(),
    ];

    $element['proc_trigger_format'] = [
      '#type' => 'radios',
      '#title' => $this->t('Trigger format'),
      '#default_value' => $this->getSetting('proc_trigger_format') ?? '',
      '#options' => $this->getTriggerFormats(),
    ];

    $element['proc_master_field_submit_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Submit element ID for the master field'),
      '#description' => $this->t('Define the submit element ID of the form where this field is located. Example: edit-submit'),
      '#default_value' => $this->getSetting('proc_master_field_submit_id') ?? '',
      '#disabled' => FALSE,
    ];

    $element['proc_decryption_trigger_fields'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Trigger decryption of another encrypted field'),
      '#default_value' => $this->getSetting('proc_decryption_trigger_fields') ?? '',
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();

    $operators = $this->getMatchOperatorOptions();
    $summary[] = $this->t('Autocomplete matching: @match_operator', ['@match_operator' => $operators[$this->getSetting('match_operator')]]);
    $size = $this->getSetting('match_limit') ?: $this->t('unlimited');
    $summary[] = $this->t('Autocomplete suggestion list size: @size', ['@size' => $size]);
    $summary[] = $this->t('Textfield size: @size', ['@size' => $this->getSetting('size')]);
    $placeholder = $this->getSetting('placeholder');
    $summary[] = (!empty($placeholder))
      ? $this->t('Placeholder: @placeholder', ['@placeholder' => $placeholder])
      : $this->t('No placeholder');
    $proc_recipients_to_field = $this->getSetting('proc_field_recipients_to_field');
    $summary[] = $this->t('Machine name of recipients To: @proc_field_recipients_to_field' , ['@proc_field_recipients_to_field' => $proc_recipients_to_field]);
    $proc_recipients_cc_field = $this->getSetting('proc_field_recipients_cc_field');
    $summary[] = $this->t('Machine name of recipients CC: @proc_field_recipients_cc_field' , ['@proc_field_recipients_cc_field' => $proc_recipients_cc_field]);
    $proc_field_recipients_manual_fetcher = $this->getSetting('proc_field_recipients_manual_fetcher');
    $summary[] = $this->t('Default recipients: @proc_field_recipients_manual_fetcher' , ['@proc_field_recipients_manual_fetcher' => $proc_field_recipients_manual_fetcher]);
    $proc_field_recipients_fetcher_endpoint = $this->getSetting('proc_field_recipients_fetcher_endpoint');
    $summary[] = $this->t('Recipients fetcher endpoint: @proc_field_recipients_fetcher_endpoint' , ['@proc_field_recipients_fetcher_endpoint' => $proc_field_recipients_fetcher_endpoint]);
    $proc_field_recipients_fetcher_endpoint_filter = $this->getSetting('proc_field_recipients_fetcher_endpoint_filter');
    $summary[] = $this->t('Machine name of fetcher filter: @proc_field_recipients_fetcher_endpoint_filter', [
      '@proc_field_recipients_fetcher_endpoint_filter' => ($proc_field_recipients_fetcher_endpoint_filter)
        ? $this->getFetcherEndpointFilter()[$proc_field_recipients_fetcher_endpoint_filter]
        : '',
    ]);
    $proc_field_recipients_fetcher_endpoint_filter_type = $this->getSetting('proc_field_recipients_fetcher_endpoint_filter_type');
    $summary[] = $this->t('Recipients fetcher type: @proc_field_recipients_fetcher_endpoint_filter_type', [
      '@proc_field_recipients_fetcher_endpoint_filter_type' => ($proc_field_recipients_fetcher_endpoint_filter_type)
        ? $this->getFetcherEndpointFilterTypes()[$proc_field_recipients_fetcher_endpoint_filter_type]
        : '',
    ]);
    $proc_field_mode = $this->getSetting('proc_field_mode');
    $summary[] = $this->t('Operation: @proc_field_mode', [
      '@proc_field_mode' => ($proc_field_mode)
        ? $this->getFieldOperationModes()[$proc_field_mode]
        : '',
    ]);
    $proc_field_input_mode = $this->getSetting('proc_field_input_mode');
    $summary[] = $this->t('Input mode: @proc_field_input_mode', [
      '@proc_field_input_mode' => ($proc_field_input_mode)
        ? $this->getInputModes()[$proc_field_input_mode]
        : '',
    ]);
    $proc_field_decryption_mode = $this->getSetting('proc_field_decryption_mode');
    $summary[] = $this->t('Decryption mode: @proc_field_decryption_mode', [
      '@proc_field_decryption_mode' => ($proc_field_decryption_mode)
        ? $this->getDecryptionModes()[$proc_field_decryption_mode]
        : '',
    ]);
    $proc_allow_password_caching = $this->getSetting('proc_allow_password_caching');
    $summary[] = $this->t('Password caching: @proc_allow_password_caching', [
      '@proc_allow_password_caching' => ($proc_allow_password_caching)
        ? $this->getPasswordCachingOptions()[$proc_allow_password_caching]
        : '',
    ]);
    $proc_trigger_format = $this->getSetting('proc_trigger_format');
    $summary[] = $this->t('Trigger format: @proc_trigger_format', [
      '@proc_trigger_format' => ($proc_trigger_format)
        ? $this->getTriggerFormats()[$proc_trigger_format]
        : '',
    ]);
    $proc_master_field_submit_id = $this->getSetting('proc_master_field_submit_id');
    $summary[] = $this->t('Submit element ID: @proc_master_field_submit_id', ['@proc_master_field_submit_id' => $proc_master_field_submit_id]);
    $proc_decryption_trigger_fields = $this->getSetting('proc_decryption_trigger_fields');
    $summary[] = $this->t('Trigger other fields on decryption: @proc_decryption_trigger_fields', [
      '@proc_decryption_trigger_fields' => ($proc_decryption_trigger_fields)
        ? $this->t('Yes')
        : $this->t('No')
    ]);

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    // Call the parent formElement method to get the base element.
    $element = parent::formElement($items, $delta, $element, $form, $form_state);
    // Retrieve element forms state value.
    $element_value = $form_state->getValue($items->getName());

    // Extract widget settings into local variables.
    // @TODO: Move these to class properties.
    $element['#proc-settings']['field_input_mode'] = $this->getSetting('proc_field_input_mode') ?? '';
    $element['#proc-settings']['to_recipients_field_name'] = $this->getSetting('proc_field_recipients_to_field') ?? '';
    $element['#proc-settings']['carbon_copy_recipients_field_name'] = $this->getSetting('proc_field_recipients_cc_field') ?? '';
    $element['#proc-settings']['preset_recipients_csv'] = $this->getSetting('proc_field_recipients_manual_fetcher') ?? '';
    $element['#proc-settings']['fetcher_endpoint'] = $this->getSetting('proc_field_recipients_fetcher_endpoint') ?? '';
    $element['#proc-settings']['fetcher_endpoint_filter'] = $this->getSetting('proc_field_recipients_fetcher_endpoint_filter') ?? '';
    $element['#proc-settings']['fetcher_endpoint_filter_type'] = $this->getSetting('proc_field_recipients_fetcher_endpoint_filter_type') ?? '';
    $element['#proc-settings']['cache_password'] = $this->getSetting('proc_allow_password_caching') ?? '0';
    $element['#proc-settings']['trigger_format'] = $this->getSetting('proc_trigger_format') ?? '0';
    $element['#proc-settings']['submit_element_id'] = $this->getSetting('proc_master_field_submit_id') ?? '';
    $element['#proc-settings']['trigger_fields'] = $this->getSetting('proc_decryption_trigger_fields') ?? '0';
    $element['#proc-settings']['proc_field_decryption_mode'] = $this->getSetting('proc_field_decryption_mode') ?? '';
    $element['#proc-settings']['proc_field_name'] = $items->getName();
    $element['#proc-settings']['proc_field_mode'] = $this->getSetting('proc_field_mode') ?? 0;

    if ($element['#proc-settings']['proc_field_mode'] !== "0") {
      $element['#type'] = 'procfield';
    }

    // Keep form state value if element isn't encrypted yet.
    if (
      empty($element['target_id']['#default_value'])
      && empty($element['target_id']['#attributes']['value'])
      && !empty($element_value)
      ) {
      $element['target_id']['#attributes']['value'] = $element_value[$delta]['target_id'];
    }

    $element['#element_validate'][] = [$this, 'validate'];

    $procElement = new ProcField($element, $items->getName(), $items->getEntity()->getEntityTypeId());
    $event = new ProcElementEvent($procElement);
    \Drupal::service('event_dispatcher')->dispatch($event, ProcEvents::PROC_ELEMENT_SETTINGS);

    return $element;
  }

  /**
   * Return field operation mode options.
   */
  private function getFieldOperationModes() {
    return [
      0 => $this->t('Disabled'),
      1 => $this->t('Only encryption'),
      2 => $this->t('Encryption and signature'),
    ];
  }

  /**
   * Return field input mode options.
   */
  private function getInputModes() {
    return [
      0 => $this->t('File'),
      1 => $this->t('Text area'),
      2 => $this->t('Text field'),
      3 => $this->t('Date')
    ];
  }

  /**
   * Return decryption mode options.
   */
  private function getDecryptionModes() {
    return [
      0 => $this->t('New page'),
      1 => $this->t('Dialog'),
    ];
  }

  /**
   * Return password caching options.
   */
  private function getPasswordCachingOptions() {
    return [
        0 => $this->t('Disable password caching'),
        1 => $this->t('Show checkbox for allowing password caching'),
        2 => $this->t('Hide the checkbox and apply password caching by default'),
    ];
  }

  /**
   * Return trigger format options.
   */
  private function getTriggerFormats() {
    return [
      0 => $this->t('Encrypt and Decrypt buttons'),
      1 => $this->t('Toggle checkbox (check for encryption, uncheck for decryption)'),
    ];
  }

  /**
   * Return fetcher endpoint filter options.
   */
  private function getFetcherEndpointFilter() {
    return [
      0 => $this->t('Encrypt and Decrypt buttons'),
      1 => $this->t('Toggle checkbox (check for encryption, uncheck for decryption)'),
    ];
  }

  /**
   * Return fetcher endpoint filter type options.
   */
  private function getFetcherEndpointFilterTypes() {
    return [
      0 => $this->t('ID'),
      1 => $this->t('Name'),
    ];
  }

  /**
   * Remove reference if field is empty or keep it if encryption is unchanged.
   */
  public function validate($element, FormStateInterface $form_state) {
    if (empty($form_state->getValue($element['#proc-settings']['proc_field_name'])[0]['target_id'])) {
      $form_state->setValueForElement($element, [0]);
    }
    if ($form_state->getValue($element['#proc-settings']['proc_field_name'])[0]['target_id'] == '*******') {
      $form_state->setValueForElement($element, ['target_id' => $element['target_id']['#default_value']->id()]);
    }
  }
}
